package mi.hdm.recipes;

import mi.hdm.exceptions.InvalidIngredientException;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

public class IngredientManager {
    private static final Logger log = LogManager.getLogger(IngredientManager.class);

    private final Map<String, Ingredient> allIngredients; //Map key of ingredient to ingredient object

    public IngredientManager() {
        allIngredients = new HashMap<>();
    }

    public IngredientManager(List<Ingredient> ingredients) {
        allIngredients = new HashMap<>();
        ingredients.forEach(i -> allIngredients.put(i.getUniqueCode(), i));
        log.info("Created ingredient manager with {} ingredients.", allIngredients.keySet().size());
    }

    public IngredientManager(Map<String, Ingredient> ingredients) {
        allIngredients = new HashMap<>(ingredients);
        log.info("Created ingredient manager with {} ingredients.", allIngredients.keySet().size());
    }


    /**
     * Adds an ingredient if there is no equal ingredient (no ingredient with the same name).
     *
     * @param in Ingredient that is to be added
     * @return True if the ingredient has been added (meaning no equal ingredient existed before), otherwise false.
     */
    public void addIngredient(Ingredient in) {
        if (getIngredientByName(in.getName()).isPresent()) {
            log.error("Ingredient '{}' not added because it already exists.", in.getName());
            throw new InvalidIngredientException("Ingredient already exists.");
        } else {
            log.info("Ingredient '{}' added successfully.", in.getName());
            allIngredients.put(in.getUniqueCode(), in);
        }
    }

    public void addIngredient(Measurement unit, String name, NutritionTable nutritionTable) {
        Ingredient in = new Ingredient(unit, name, nutritionTable);
        addIngredient(in);
    }

    public void deleteIngredient(String name) {
        Ingredient ingredient = getIngredientByName(name).orElseThrow(() -> new InvalidIngredientException("No ingredient with name " + name));
        log.info("Ingredient '{}' deleted successfully.", name);
        allIngredients.values().remove(ingredient);
    }

    public void clearIngredients() {
        log.info("List of ingredients cleared successfully.");
        allIngredients.clear();
    }

    public Optional<Ingredient> getIngredient(String code) {
        return Optional.ofNullable(allIngredients.get(code));
    }

    public Map<String, Ingredient> getAllIngredients() {
        return allIngredients;
    }

    public Map<Ingredient, Integer> getIngredientsFromKeys(Map<String, Integer> m) {
        Map<Ingredient, Integer> result = new HashMap<>();
        for (String key : m.keySet()) {
            result.put(allIngredients.get(key), m.get(key));
        }
        return result;
    }

    private Optional<Ingredient> getIngredientByName(String name) {
        for (final Ingredient i : allIngredients.values()) {
            if (name.equals(i.getName())) {
                return Optional.of(i);
            }
        }
        return Optional.empty();
    }
}
